"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MongoDBOIDCPluginImpl = void 0;
const types_1 = require("./types");
const util_1 = require("./util");
const openid_client_1 = require("openid-client");
const rfc_8252_http_server_1 = require("./rfc-8252-http-server");
const open_1 = __importDefault(require("open"));
const util_2 = require("util");
const crypto_1 = require("crypto");
const events_1 = require("events");
const api_1 = require("./api");
const child_process_1 = require("child_process");
// eslint-disable-next-line @typescript-eslint/consistent-type-imports
let _electron = undefined;
async function getDefaultOpenBrowser() {
    // If running electron, use electron.shell.openExternal() by default
    // to open a browser.
    if (process.versions.electron && _electron !== 'cannot-require') {
        try {
            _electron ?? (_electron = await Promise.resolve().then(() => __importStar(require('electron'))));
            return ({ url }) => 
            // eslint-disable-next-line @typescript-eslint/consistent-type-imports
            _electron.shell.openExternal(url);
        }
        catch {
            _electron = 'cannot-require';
        }
    }
    // Otherwise, use open() from npm.
    return async ({ url }) => {
        const child = await (0, open_1.default)(url);
        child.unref();
        return child;
    };
}
const kEnableFallback = Symbol.for('@@mdb.oidcplugin.kEnableFallback');
/** @internal */
class MongoDBOIDCPluginImpl {
    constructor(options) {
        this.mapUserToAuthState = new Map();
        this.options = options;
        this.logger = options.logger ?? new events_1.EventEmitter();
        this.mongoClientOptions = {
            authMechanismProperties: {
                REQUEST_TOKEN_CALLBACK: this.requestToken.bind(this),
                REFRESH_TOKEN_CALLBACK: this.refreshToken.bind(this),
            },
        };
    }
    // Is this flow supported and allowed?
    isFlowAllowed(flow) {
        if (this.options.allowedFlows?.includes(flow) === false)
            return false;
        if (flow === 'auth-code' && this.options.openBrowser === false)
            return false;
        if (flow === 'device-auth' && !this.options.notifyDeviceFlow)
            return false;
        return true;
    }
    // Return the current state for a given [server, username] configuration,
    // or create a new one if none exists.
    getAuthState(serverMetadata, principalName) {
        if (!serverMetadata.clientId) {
            throw new types_1.MongoDBOIDCError('No clientId passed in server OIDC metadata object');
        }
        principalName ?? (principalName = null);
        const key = JSON.stringify({
            clientId: serverMetadata.clientId,
            principalName,
        });
        const existing = this.mapUserToAuthState.get(key);
        if (existing)
            return existing;
        const newState = {
            serverOIDCMetadata: serverMetadata,
            currentAuthAttempt: null,
            currentTokenSet: null,
        };
        this.mapUserToAuthState.set(key, newState);
        return newState;
    }
    getRedirectURI() {
        if (!this.isFlowAllowed('auth-code'))
            return '';
        // TODO(MONGOSH-1394): Properly standardize a port
        return this.options.redirectURI ?? 'http://localhost:27097/redirect';
    }
    getInitialOIDCIssuerAndClientParams(serverMetadata) {
        return {
            scope: [
                ...new Set([
                    'openid',
                    'offline_access',
                    ...(serverMetadata.requestScopes ?? []),
                ]),
            ].join(' '),
            issuerParams: {
                authorization_endpoint: serverMetadata.authorizationEndpoint,
                token_endpoint: serverMetadata.tokenEndpoint,
                device_authorization_endpoint: serverMetadata.deviceAuthorizationEndpoint,
                // Required by the TS definitions, but we just don't have this data
                // at this point. We re-define it later where necessary.
                issuer: '',
            },
            clientParams: {
                client_id: serverMetadata.clientId,
                client_secret: serverMetadata.clientSecret,
                redirect_uris: [this.getRedirectURI()],
                response_types: ['code'],
                // At least Okta requires this:
                token_endpoint_auth_method: serverMetadata.clientSecret
                    ? 'client_secret_post'
                    : 'none',
            },
        };
    }
    async openBrowser(options) {
        // Consistency check: options.url is a valid URL.
        new URL(options.url);
        this.logger.emit('mongodb-oidc-plugin:open-browser', {
            customOpener: !!this.options.openBrowser,
        });
        if (this.options.openBrowser === false) {
            // We should never really get to this point
            throw new Error('Cannot open browser if `openBrowser` is false');
        }
        if (typeof this.options.openBrowser === 'function') {
            return await this.options.openBrowser(options);
        }
        if (this.options.openBrowser === undefined) {
            const defaultOpener = await getDefaultOpenBrowser();
            return await defaultOpener(options);
        }
        if (typeof this.options.openBrowser?.command === 'string') {
            const child = (0, child_process_1.spawn)(this.options.openBrowser.command, [options.url], {
                shell: true,
                stdio: 'ignore',
                detached: true,
                signal: this.options.openBrowser.abortable ? options.signal : undefined,
            });
            child.unref();
            return child;
        }
        throw new Error('Unknown format for `openBrowser`');
    }
    async notifyDeviceFlow(deviceFlowInformation) {
        if (!this.options.notifyDeviceFlow) {
            // Should never happen.
            throw new Error('notifyDeviceFlow() requested but not provided');
        }
        this.logger.emit('mongodb-oidc-plugin:notify-device-flow');
        await this.options.notifyDeviceFlow(deviceFlowInformation);
    }
    storeTokenSet(state, tokenSet, client) {
        state.currentTokenSet = {
            set: tokenSet,
            tryRefresh: async () => {
                // Only refresh this token set if it is the one currently
                // being used.
                if (state.currentTokenSet?.set !== tokenSet)
                    return false;
                try {
                    const refreshedTokens = await client.refresh(tokenSet);
                    // Check again to avoid race conditions.
                    if (state.currentTokenSet?.set === tokenSet) {
                        this.storeTokenSet(state, refreshedTokens, client);
                        return true;
                    }
                }
                catch (err) {
                    this.logger.emit('mongodb-oidc-plugin:refresh-failed', {
                        error: (0, util_1.errorString)(err),
                    });
                }
                return false;
            },
        };
    }
    async authorizationCodeFlow(state, signal) {
        const { scope, issuerParams, clientParams } = this.getInitialOIDCIssuerAndClientParams(state.serverOIDCMetadata);
        const codeVerifier = openid_client_1.generators.codeVerifier();
        const codeChallenge = openid_client_1.generators.codeChallenge(codeVerifier);
        let issuer = new openid_client_1.Issuer(issuerParams);
        let client = new issuer.Client(clientParams);
        const oidcStateParam = (await (0, util_2.promisify)(crypto_1.randomBytes)(16)).toString('hex');
        const server = new rfc_8252_http_server_1.RFC8252HTTPServer({
            redirectUrl: this.getRedirectURI(),
            logger: this.logger,
            redirectServerRequestHandler: this.options.redirectServerRequestHandler,
            oidcStateParam,
        });
        let paramsUrl = '';
        let enableFallback = true;
        try {
            await (0, util_1.withAbortCheck)(signal, async ({ signalCheck, signalPromise }) => {
                await Promise.race([server.listen(), signalPromise]);
                const authCodeFlowUrl = client.authorizationUrl({
                    scope,
                    code_challenge: codeChallenge,
                    code_challenge_method: 'S256',
                    state: oidcStateParam,
                });
                const { localUrl, onAccessed: onLocalUrlAccessed } = await server.addRedirect(authCodeFlowUrl);
                signalCheck();
                // Handle errors from opening a browser but do not await the Promise
                // in case it only resolves when the browser exits (which is the case
                // for the default `open` handler).
                const browserStatePromise = new Promise((resolve, reject) => {
                    this.openBrowser({ url: localUrl, signal })
                        .then((browserHandle) => {
                        browserHandle?.once('error', (err) => reject(err));
                        browserHandle?.once('exit', (code) => {
                            if (code !== 0)
                                reject(new types_1.MongoDBOIDCError(`Opening browser failed with exit code ${code}`));
                        });
                    })
                        .catch(reject);
                });
                const timeout = new Promise((resolve, reject) => {
                    if (this.options.openBrowserTimeout !== 0) {
                        setTimeout(reject, this.options.openBrowserTimeout ?? api_1.kDefaultOpenBrowserTimeout, new types_1.MongoDBOIDCError('Opening browser timed out')).unref();
                    }
                });
                browserStatePromise.catch(() => {
                    /* squelch UnhandledPromiseRejectionWarning */
                });
                timeout.catch(() => {
                    /* ditto */
                });
                await Promise.race([
                    onLocalUrlAccessed,
                    timeout,
                    browserStatePromise,
                    signalPromise,
                ]);
                // If we reached this point, we know that we have successfully opened
                // a server listening on a local port and a browser and that the
                // browser accessed the server. We do not want to fall back to another
                // flow anymore, any error from here on is most likely a genuine
                // authentication error.
                enableFallback = false;
                paramsUrl = await server.waitForOIDCParamsAndClose({ signal });
            });
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
        }
        catch (err) {
            // Tell the outer logic here to fallback to device auth flow if it is
            // available if any of the steps above failed.
            if (Object.isExtensible(err)) {
                err[kEnableFallback] = enableFallback;
            }
            throw err;
        }
        finally {
            await server.close();
        }
        const params = client.callbackParams(paramsUrl);
        // The oidc-client library requires `issuer` to be set here; we did not
        // set it when we assembled the original `issuerParams` (because it was
        // not available as information), but we should have received it from the
        // callback parameters and can set it here.
        issuerParams.issuer = String(params.iss);
        issuer = new openid_client_1.Issuer(issuerParams);
        client = new issuer.Client(clientParams);
        client.validateIdToken = () => {
            // Do not attempt to validate the ID token. The client library would
            // do this by default, but we do not have access to the necessary
            // JWKS here. This is in direct conflict with the OIDC spec (!!)
            // https://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation:
            //
            // > Clients MUST validate the ID Token in the Token Response in the following manner: [...]
            //
            // We accept this because for the purposes of this plugin, we do not
            // interact with the identity token, nor do we actually consume the
            // access token. We leave it to the the server to validate that the token is
            // valid, refers to the right identity, and comes from the right source.
        };
        const tokenSet = await client.callback(this.getRedirectURI(), params, {
            code_verifier: codeVerifier,
            state: oidcStateParam,
        });
        this.storeTokenSet(state, tokenSet, client);
    }
    async deviceAuthorizationFlow(state, signal) {
        const { scope, issuerParams, clientParams } = this.getInitialOIDCIssuerAndClientParams(state.serverOIDCMetadata);
        const issuer = new openid_client_1.Issuer(issuerParams);
        const client = new issuer.Client(clientParams);
        client.validateIdToken = () => {
            /* see above */
        };
        await (0, util_1.withAbortCheck)(signal, async ({ signalCheck, signalPromise }) => {
            const deviceFlowHandle = await Promise.race([
                client.deviceAuthorization({
                    client_id: clientParams.client_id,
                    scope,
                }),
                signalPromise,
            ]);
            signalCheck();
            await this.notifyDeviceFlow({
                userCode: deviceFlowHandle.user_code,
                verificationUrl: deviceFlowHandle.verification_uri,
            });
            const tokenSet = await deviceFlowHandle.poll({ signal });
            this.storeTokenSet(state, tokenSet, client);
        });
    }
    async initiateAuthAttempt(state, driverAbortSignal) {
        (0, util_1.throwIfAborted)(this.options.signal);
        (0, util_1.throwIfAborted)(driverAbortSignal);
        const combinedAbortController = new util_1.AbortController();
        const optionsAbortCb = () => {
            // @ts-expect-error TS doesn't understand .abort(reason)
            combinedAbortController.abort(this.options.signal.reason);
        };
        const driverAbortCb = () => {
            // @ts-expect-error TS doesn't understand .abort(reason)
            combinedAbortController.abort(driverAbortSignal.reason);
        };
        this.options.signal?.addEventListener('abort', optionsAbortCb);
        driverAbortSignal?.addEventListener('abort', driverAbortCb);
        const signal = combinedAbortController.signal;
        try {
            get_tokens: {
                if ((state.currentTokenSet?.set?.expires_in ?? 0) > 5 * 60) {
                    this.logger.emit('mongodb-oidc-plugin:skip-auth-attempt', {
                        reason: 'not-expired',
                    });
                    break get_tokens;
                }
                if (await state.currentTokenSet?.tryRefresh?.()) {
                    this.logger.emit('mongodb-oidc-plugin:skip-auth-attempt', {
                        reason: 'refresh-succeeded',
                    });
                    break get_tokens;
                }
                state.currentTokenSet = null;
                let error;
                if (this.isFlowAllowed('auth-code')) {
                    try {
                        this.logger.emit('mongodb-oidc-plugin:auth-attempt-started', {
                            flow: 'auth-code',
                        });
                        await this.authorizationCodeFlow(state, signal);
                        this.logger.emit('mongodb-oidc-plugin:auth-attempt-succeeded');
                        break get_tokens;
                    }
                    catch (err) {
                        error = err;
                        this.logger.emit('mongodb-oidc-plugin:auth-attempt-failed', {
                            error: (0, util_1.errorString)(err),
                        });
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any
                        if (!err?.[kEnableFallback])
                            throw err;
                    }
                }
                if (this.isFlowAllowed('device-auth')) {
                    try {
                        this.logger.emit('mongodb-oidc-plugin:auth-attempt-started', {
                            flow: 'device-auth',
                        });
                        await this.deviceAuthorizationFlow(state, signal);
                        this.logger.emit('mongodb-oidc-plugin:auth-attempt-succeeded');
                        break get_tokens;
                    }
                    catch (err) {
                        this.logger.emit('mongodb-oidc-plugin:auth-attempt-failed', {
                            error: (0, util_1.errorString)(err),
                        });
                        throw err;
                    }
                }
                if (error)
                    throw error;
            }
            if (!state.currentTokenSet?.set?.access_token) {
                throw new types_1.MongoDBOIDCError('Could not retrieve valid access token');
            }
        }
        catch (err) {
            this.logger.emit('mongodb-oidc-plugin:auth-failed', {
                error: (0, util_1.errorString)(err),
            });
            throw err;
        }
        finally {
            this.options.signal?.removeEventListener('abort', optionsAbortCb);
            driverAbortSignal?.removeEventListener('abort', driverAbortCb);
        }
        this.logger.emit('mongodb-oidc-plugin:auth-succeeded', {
            hasRefreshToken: !!state.currentTokenSet.set.refresh_token,
            expiresAt: state.currentTokenSet.set.expires_at
                ? new Date(state.currentTokenSet.set.expires_at * 1000).toISOString()
                : null,
        });
        return {
            accessToken: state.currentTokenSet.set.access_token,
            refreshToken: state.currentTokenSet.set.refresh_token,
            // Passing `expiresInSeconds: 0` results in the driver not caching the token.
            // We perform our own caching here inside the plugin, so interactions with the
            // cache of the driver are not really required or necessarily helpful.
            // The driver cache has a finer cache key (host address instead of clientId),
            // so may require more authentication attempts, and is global,
            // not per-MongoClient.
            // It probably would be fine to pass in the actual expiration time here, but
            // there seem to be no benefits to doing so.
            expiresInSeconds: 0,
        };
    }
    async requestToken(principalName, serverMetadata, driverAbortSignal) {
        const state = this.getAuthState(serverMetadata, principalName);
        if (state.currentAuthAttempt) {
            return await state.currentAuthAttempt;
        }
        // The currently plan is for the 6.x driver (which may drop support
        // for Node.js 14.x) to pass in an actual AbortSignal here. For
        // compatibility with the 5.x driver/AbortSignal-less-Node.js, we accept
        // a timeout in milliseconds as well.
        if (typeof driverAbortSignal === 'number') {
            driverAbortSignal = (0, util_1.timeoutSignal)(driverAbortSignal);
        }
        const newAuthAttempt = this.initiateAuthAttempt(state, driverAbortSignal);
        state.currentAuthAttempt = newAuthAttempt;
        newAuthAttempt.finally(() => {
            if (state.currentAuthAttempt === newAuthAttempt)
                state.currentAuthAttempt = null;
        });
        return newAuthAttempt;
    }
    async refreshToken(principalName, serverMetadata, 
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    previousResult, driverAbortSignal) {
        return await this.requestToken(principalName, serverMetadata, driverAbortSignal);
    }
}
exports.MongoDBOIDCPluginImpl = MongoDBOIDCPluginImpl;
//# sourceMappingURL=plugin.js.map