"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.timeoutSignal = exports.AbortSignal = exports.AbortController = exports.errorString = exports.withAbortCheck = exports.throwIfAborted = void 0;
class AbortError extends Error {
    constructor() {
        super('The operation was aborted');
    }
}
function throwIfAborted(signal) {
    if (signal?.aborted)
        throw signal.reason ?? new AbortError();
}
exports.throwIfAborted = throwIfAborted;
async function withAbortCheck(signal, fn) {
    const signalCheck = () => throwIfAborted(signal);
    let reject;
    const signalPromise = new Promise((resolve, rej) => {
        reject = rej;
    });
    function listener() {
        reject(signal?.reason ?? new AbortError());
    }
    signalPromise.catch(() => {
        /* squelch UnhandledPromiseRejectionWarning */
    });
    signalCheck();
    signal?.addEventListener('abort', listener, { once: true });
    try {
        return await fn({ signalCheck, signalPromise });
    }
    finally {
        signal?.removeEventListener('abort', listener);
    }
}
exports.withAbortCheck = withAbortCheck;
function errorString(err) {
    return String(typeof err === 'object' && err && 'message' in err ? err.message : err);
}
exports.errorString = errorString;
exports.AbortController = 
// eslint-disable-next-line @typescript-eslint/no-var-requires
globalThis.AbortController ?? require('abort-controller').AbortController;
exports.AbortSignal = 
// eslint-disable-next-line @typescript-eslint/no-var-requires
globalThis.AbortSignal ?? require('abort-controller').AbortSignal;
// AbortSignal.timeout, but consistently .unref()ed
function timeoutSignal(ms) {
    const controller = new exports.AbortController();
    setTimeout(() => controller.abort(), ms).unref();
    return controller.signal;
}
exports.timeoutSignal = timeoutSignal;
//# sourceMappingURL=util.js.map